
// This software is written by HKJ from lygte-info.dk
// It is desiged for two different hardware configurations: 1) 3 power mos outputs, 2) 4 logic 5V or 3.3V output.
// They are mostly the same, but option 2) adds support for 1 extra output (4 outputs) for relay and pattern commands
// The logical version also support serial output protocols, this is not connected on the relay version

#include <EEPROM.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include "CmdProcessor.h"

// Support 1-3 mosfets with lots of functions, for digital output it can support 4 outputs, but the extra output only works for relay and pattern commands
/*
  Commands:
  RELAYS onmask  ; Bit0=Relay1, bit1=Relay1
  RELAYS?        ; Return a onmask
  RELAYON no     ; Turn one relay on, no can be from 1 to 4
  RELAYOFF no    ; Turn one relay off, no can be from 1 to 4
  RELAY? no      ; Request status for one relay, no can be from 1 to 4
  BUTTONS?       ; Return a mask of pressed buttons
  PWM no duty_cycle ; Specify pwm duty cucle
  PWMW no ms    ; Specifify pwm high time instead of duty cycle
  PWM?          ; List duty cycle for all 3 channels
  PWMFREQ frequency ; Frequency for PWM and PPPWM up to 100kHz
  PWMPERIOD ms   ; 1/PWMFREQ
  PPPWM duty_cycle  ; Push pull PWM with specified duty cycle and PWM Frequency on output 1 and 2 there is a dead time between the two outputs, output 3 cannot be used while this command is active
  PPPWMW ms    ; Specifify pwm high time instead of duty cycle
  DEADTIME us ; set from 0 to 2.5us deadtime for PPPWM, this value is saved in EEPROM
  PWMFREQ?    ; List pwm frequency
  PULSE no ontime ; Pulse one output on for specified milliseconds, only one output can be active at a time
  PULSES ontime offtime pulses  ; Send specified number of pulses to output 1
  PULSEALL time1 time2 time3        :
  SEQ time1 time2 time3 ; Turn output 1 on at time1, output 2 at time 2 and output 3 at time 3
  PATTERN time pattern  ;Output pattern once 1=relay1, 2=relay=2, 4=relay3, 7=relay1+2+3
  REPPATTERN time pattern; Repeat pattern .
  PATTERN?    ; Show current pattern
  SERIAL baud format  ;; Support from 150baud to 1M baud
  SERIAL?     ; Show current format
  TX data     ; Transmit serial data, data can be quoted strings, hex numbers, decimal numbers and color letters, used for serial, SPI and PIXEL
  REPTX data  ; Continuous transmission
  TX?         ; Show last string
  SPI rateFormat
              ; Format l=lsb first (default msb first)
              ; Format e=Select active low (Default high)
              ; Format setup data edge: .=falling, c=falling, d=raising, c&d=raising
  SPI?
  PIXEL n format ; Switch to led mode, n can specify number of pixels to update, format is led type or RGB, GRB, RGBW
  RESOLUTION? ; Returns actual counter value, this shows how good resulution is used.


  OFF           ; Turn all outputs off, same as "RELAYS 0"
  WAIT milliseconds ; Wait specified number of ms, used together with program, range is 1 to 100000
  RUN         ; Run program
  PROGRAM list of commands with | delimeter
  PROGRAM?    ; List stored program
  PROGRAMADD list of commands with | delimeter ; Add more commands to PROGRAM
  PA        ; Same as PROGRAMADD
  SELECTPGM  ; Ask for user input, then search forward for PGMx markers
  PGMx      ; Labels and stop points, used together with SELECTPGM

  no=0 means all for PWM and PULSE
  PWM, PULSE, PULSEALL, SEQ, PPPWM all runs asynchronous, i.e.they do not block the command loop. This means WAIT must be used in programs.
  PULSE, PULSEALL, SEQ, PPPWM: only one command can be activate at a time.
  PULSE, PULSEALL, SEQ: Output 1 & 2 share a timer and will use same range, 3 has separate timer.
  PPPWM  Dead time is up to 2.5usec -> 1/48M*8*15


*/

// MOS outputs are active high
// Led outputs are active high
// Button inputs are activated low

// The definition for these register bits are missing
#define UCPHA1 1
#define UDORD1 2


#define mos1SerialPin 1 // PD3 used for transmitting serial data, must be connected PC6 pin
#define mos1MOSIPin 16
#define mos3SCKPin 15

#define mos1Pin 9   // TC1 A PWM output PB5 or TC4 !B
#define mos2Pin 10  // TC1 B PWM output PB6 OR TC4 B
#define mos3Pin 5   // TC3 A PWM output PC6
#define mos4Pin 14  // PB3 Extra output without much functionality


#define led1Pin 2 // PD1
#define led2Pin 3 // PD0
#define led3Pin 4 // PD4
#define led5Pin 6 // PD7

// Buttons is read in either Intr4 or Intr3 routine with direct port access, this takes less then 2usec, but means the pins cannot be swapped in software
#define button1Pin A3   // PF4
#define button2Pin A2   // PF5
#define button3Pin A1   // PF6
#define button5Pin A0   // PF7        // This is still assumed to be bit 3 in the button status register



#define debounceCount 5   // Each tick is 4ms


// EEPROM and size constants
#define MAX_SERIAL_SIZE 128
#define MAX_PATTERN_SIZE 64
#define MAX_PROGRAM_SIZE 512
#define PROGRAM_LOCATION 10 // Uses 512 bytes
#define DEADTIME_LOCATION 9 // Used 1 byte with time in 0.01usec units.

// Mode constants
#define MODE_RELAY 0
#define MODE_PWM 1
#define MODE_PULSE 2
#define MODE_PULSES 3
#define MODE_PULSEALL 4
#define MODE_SEQ 5
#define MODE_PPPWM 6
#define MODE_PATTERN 7
#define MODE_SERIAL 8
#define MODE_SPI 9
#define MODE_PIXEL 11

#define MODE_SLOW_LIMIT 11    // Any mode above this value will call the slow handling routine

#define MODE_SLOW_PULSE 12
#define MODE_SLOW_PULSES 13
#define MODE_SLOW_PULSEALL 14
#define MODE_SLOW_SEQ 15
#define MODE_SLOW_PATTERN 17

#define PWM1_TIME 0x01      // Flags  used in timeFlag byte to distinguish between time/%/hz
#define PWM2_TIME 0x02
#define PWM3_TIME 0x04
#define FREQ_TIME 0x10

const char *paramDelimeters = ";, ";
const char *cmdDelimeter = " ";

// Structure used to store timer settings from calculation routine
struct TimerSettings {
  byte maskb1;      // Divide mask for timer 1
  byte maskb3;      // Divide mask for timer 3
  int fac1;         // Divide factor for timer 1
  int fac3;         // Divide factor for timer 3
  unsigned int o1;  // Count for time 1
  unsigned int o2;  // Count for time 2
  unsigned int o3;  // Count for time 3
};

Stream *debug = NULL;      // This stream is used for debug (And repeated RUN), actual debug code is usual commented out.

volatile boolean buttonPressed = false;
boolean inProgram = false;    // Running a program
boolean loopProgram = false;  // Request to automatic restart RUN command when finished
boolean fastpwm = false;  // When true leds will light permanently with any duty cycle setting above 0

byte mosStatus = 0;             // MOS relay output bits
volatile byte buttonStatus = 0;          // Actual button status
byte debounce = debounceCount;  // Debounce for buttons

byte ledPulse[4] = {255, 255, 255, 255}; // With this initialization the leds will flash when starting the processor
byte findPgm = 0;             // 1=Abort, '1','2','3': Search for PGM1, PGM2, PGM3

char program[512];              // Program buffer, used with PROGRAM and RUN commands
int resolution = 0;             // Current timer resolution, only used for RESOLUTION? command

volatile boolean repeatPattern = false;  // Run pattern once or repeat
volatile byte npattern = 0;              // Entries in pattern storage
volatile byte patternIdx = 0;            // Index used in interupt routine
volatile byte pattern[MAX_PATTERN_SIZE]; // Storage for patterns

byte deadtime = 250;
long pwmDutyCycle1 = 0;    // Duty cycle parameters, see timeflag for unit
long pwmDutyCycle2 = 0;
long pwmDutyCycle3 = 0;
long pwmFrequency = 10000; // Frequency setting for PWM, see timeflag for unit
long pwmLastFreq = 0;     // Used to check if frequency is changed
byte timeFlag = 0;          // Is PWM/FREQUENCY specificed in %/Hz/ms

int pulseNo = 0;            // Pulse command parameters
unsigned long pulseTime = 0;
unsigned long pulseCount = 0;
unsigned long pulseTimeOff = 0;

unsigned long pulseT1 = 0;   // pulseall and seq command parameters
unsigned long pulseT2 = 0;
unsigned long pulseT3 = 0;

long baudrate = 9600;
byte serialFormat = (byte) 0x13;         // Format bit567=no,space,mark,odd,even, bit4=1-2 stop bit, bit0-2:5-8 data bit
volatile byte serialBuffer[MAX_SERIAL_SIZE];
volatile byte nSerialBuffer = 0;
volatile byte serialBufferIdx = 0;
volatile byte serialBufferParityMask = 0; // Bit 0=8 bit mark parity, bit 1=not pariry but command, bit567=mark parity
volatile boolean repeatSerialBuffer = false;
long spirate = 100000; // Default is 100kHz
byte spiFormat = 0; // Bit 0=Clock polarity, 1=Data polarity, 2=LSB/MSB first, , 3=Enable polarity, 7=Enable SPI
byte pixelFormat = 0; // Bit 7=enabled pixel handling
unsigned int pixelCount = 0; // Number of pixels, when 0 transmit data from buffer, when not 0 repeat data from buffer

long slowUpdateInterval = 0;// Used for slow event with regular update interval, i.e. PATTERN commands
unsigned long nextSlowEvent = 0;     // Compared to millis(), when above the next step in slow processing must be handled
unsigned long slowTimer = 0; // Used for timing slow events, this is the time since start of the slow command

int mode = MODE_RELAY;      // Mode used for last command

CmdProcessor cmdProcessor;

//----------------------------------------------------------------------------------------------------
// Initial configutation
void setup() {
  *program = 0;
  *serialBuffer = 0;
  *pattern = 0;
  cmdProcessor.init("SS-Relay,1,1.0", handleCmds);

  deadtime = EEPROM.read(DEADTIME_LOCATION);

  // Enable pullup on pins
  pinMode(button1Pin, INPUT);
  pinMode(button2Pin, INPUT);
  pinMode(button3Pin, INPUT);
  pinMode(button5Pin, INPUT);
  digitalWrite(button1Pin, HIGH);
  digitalWrite(button2Pin, HIGH);
  digitalWrite(button3Pin, HIGH);
  digitalWrite(button5Pin, HIGH);

  // Set relay pins as output and secure correct status on all pins
  mode = MODE_SEQ; // Forces clearOutputs() to do a full timer reset
  setRelays();

  pinMode(mos1SerialPin, INPUT);
  pinMode(mos1MOSIPin, INPUT);
  pinMode(mos3SCKPin, INPUT);

  pinMode(mos1Pin, OUTPUT);
  pinMode(mos2Pin, OUTPUT);
  pinMode(mos3Pin, OUTPUT);
  pinMode(mos4Pin, OUTPUT);
  pinMode(led1Pin, OUTPUT);
  pinMode(led2Pin, OUTPUT);
  pinMode(led3Pin, OUTPUT);
  pinMode(led5Pin, OUTPUT);

  // Setup timer counter 4 for interrupts, this is used for scanning inputs
  TCCR4B = (1 << CS43);    // 128 prescaler 250Khz or 4ms intr
  TIMSK4 = (1 << TOIE4);   // enable timer overflow interrupt
}

// Storing commands this way saves ram
PROGMEM const char cmdList[] = "RELAYS\0RELAYON\0RELAYOFF\0PWMFREQ\0PWMPERIOD\0PWM\0PWMW\0PPPWM\0PPPWMW\0PULSE\0PULSEALL\0"
                               "SEQ\0SEQUENCE\0DEADTIME\0PATTERN\0REPPATTERN\0OFF\0WAIT\0PROGRAM\0PROGRAMADD\0PA\0RUN\0SERVO\0"
                               "SERIAL\0SPI\0PIXEL\0TX\0REPTX\0SELECTPGM\0PULSES\0\0";
PROGMEM const char cmdQList[] = "RELAYS?\0RELAY?\0PWMFREQ?\0PWM?\0PROGRAM?\0DEADTIME?\0PATTERN?\0RESOLUTION?\0SERIAL?\0SPI?\0TX?\0"
                                "PIXEL?\0MODE?\0PULSE?\0R?\0PULSES?\0\0"

                                ;
//----------------------------------------------------------------------------------------------------
void handleCmds(Stream *stream, char* cmd) {
  debug = stream;
  char *p = cmd; while (*p != 0) {
    *p = toupper(*p);  // Convert command to upper case
    p++;
  }
  if (strncmp(cmd, "PGM", 3) == 0) {
    if (findPgm > '0') {
      //This is a search for the program
      char c = *(cmd + 3); // Fetch label number
      if (c == findPgm) {
        findPgm = 0;    // Disable searching, next command will be processed
      }
    } else {    // Not in a search, abort program
      findPgm = 1;
    }
    return;   // All processing for PGM is done
  }
  if (stream != NULL && strcmp(cmd, "BUTTONS?") == 0) cmdButtonsQ(stream);  // This command must be processed before the key break function
  if ((buttonPressed && (buttonStatus != 0)) || findPgm > 0) {
    loopProgram = false;
    return; // Skip command processing when a key is pressed or searching for PGM label.
  }
  if (!inProgram) loopProgram = false;
  switch (cmdidx(cmd, cmdList)) {
    case 1: cmdRelays(); break; // RELAYS
    case 2: cmdRelayOn(); break; // RELAYON
    case 3: cmdRelayOff(); // RELAYOFF
    case 4: cmdPWMFreq(); break; // PWMFREQ
    case 5: cmdPWMPeriod(); break; // PWMPERIOD
    case 6: cmdPWM(); break; // PWM
    case 7: cmdPWMW(); break; // PWMW
    case 8: cmdPushPullDrive(); break; // PPPWM
    case 9: cmdPushPullDriveW(); break; // PPPWMW
    case 10: cmdPulse(); break; //PULSE
    case 11: cmdPulseAll(); break; // PULSEALL
    case 12: cmdSequence(); break; // SEQ
    case 13: cmdSequence(); break; // SEQUENCE
    case 14: cmdDeadTime(); break; // DEADTIME
    case 15: cmdPattern(false); break; // PATTERN
    case 16: cmdPattern(true); break; // REPPATTERN
    case 17: cmdOff(); break; // OFF
    case 18: cmdWait(); break; // WAIT
    case 19: cmdProgram(); break; // PROGRAM
    case 20: cmdProgramAdd(); break; // PROGRAMADD
    case 21: cmdProgramAdd(); break; // PA
    case 22: cmdRun(stream); break; // RUN
    case 23: cmdServo(); break; // SERVO
    case 24: cmdSerial(); break; // SERIAL
    case 25: cmdSPI(); break; // SPI
    case 26: cmdPixel(); break; // PIXEL
    case 27: cmdTx(false); break; // TX
    case 28: cmdTx(true); break; // REPTX
    case 29: if (inProgram) cmdSelect(); break; // SELECTPGM
    case 30: cmdPulses(); break; // PULSES
    default:
      if (stream != NULL) {  // When starting programs with the button stream will be NULL, this way I ignore commands using the stream.
        switch (cmdidx(cmd, cmdQList)) {
          case 1 : cmdRelaysQ(stream); break; // RELAYS?
          case 2 : cmdRelayQ(stream); break; // RELAY?
          case 3 : cmdPWMFreqQ(stream); break; // PWMFREQ?
          case 4 : cmdPWMQ(stream); break; // PWM?
          case 5 : cmdProgramQ(stream); break; // PROGRAM?
          case 6 : cmdDeadTimeQ(stream); break; // DEADTIME?
          case 7 : cmdPatternQ(stream); break; // PATTERN?
          case 8 : cmdResolutionQ(stream); break; // RESOLUTION?
          case 9 : cmdSerialQ(stream); break; // SERIAL?
          case 10 : cmdSPIQ(stream); break; // SPI?
          case 11 : cmdTxQ(stream); break; // TX?
          case 12 : cmdPixelQ(stream); break; // PIXEL?
          case 13 : cmdModeQ(stream); break; // MODE?
          case 14 : cmdPulseQ(stream); break; // PULSE?
          case 15 : cmdRegisterQ(stream); break; // R?
          case 16 : cmdPulseQ(stream); break; // PULSES?
        }
      }
  }
}

//----------------------------------------------------------------------------------------------------
// Reset all modes and prepare for actual mode configuration
void clearOutputs() {
  if (mode != MODE_RELAY) {
    resolution = 0;
    // Stop timers
    TCCR1B = 0;
    TCCR3B = 0;
    fastpwm = false;

    // Reset timer outputs, this prevents spurious outputs next time they are activated
    TCCR1A = (1 << COM1A1) | (1 << COM1B1);   // Force outputs low on match
    TCCR3A = (1 << COM3A1);
    TCCR1C = (1 << FOC1A) | (1 << FOC1B); // Reset all outputs
    TCCR3C = (1 << FOC3A);

    // Disconnect them from output
    TCCR1A = 0;
    TCCR3A = 0;

    // Set digital outputs to zero.
    pinMode(mos1SerialPin, INPUT);
    pinMode(mos1MOSIPin, INPUT);
    pinMode(mos3SCKPin, INPUT);
    pinMode(mos1Pin, OUTPUT);
    pinMode(mos3Pin, OUTPUT);
  }
  if (mode == MODE_PPPWM) {
    TIMSK3 = 0;             // No interrupt from Timer3
    PLLFRQ &= ~((1 << PLLTM1) | (1 << PLLTM0));  // Remove internal clock boost to timer 4
    // Switch back to timer 4 as intr source
    TCCR4A = 0;
    TCCR4B = (1 << CS43);    // 128 prescaler 250Khz or 4ms intr
    TCCR4D = 0;
    TIMSK4 = (1 << TOIE4);   // enable timer overflow interrupt
    pwmDutyCycle2 = 0;
    pwmDutyCycle3 = 0;
  }
  if (mode == MODE_PWM) {
    pwmDutyCycle1 = 0;
    pwmDutyCycle2 = 0;
    pwmDutyCycle3 = 0;
  }
  if (mode == MODE_PATTERN) {
    // Disable interrupt from timer 1
    TCCR1A = 0;
    TIMSK1 = 0;             // No interrupt from Timer1
  }

  if (mode == MODE_SERIAL) {
    UCSR1A = 0;
    UCSR1B = 0;         // Disable UART
    UCSR1C = 0;
  }
  if (mode == MODE_SPI) {
    SPSR = 0;         // Disable SPI
  }

  PORTB &= ~(0x64);   // Output bits off
  PORTC &= ~(0x40);
  pwmLastFreq = 0;    // Force reprogramming of divider and top value in PWM modes
}


void sendNeoPixelData(unsigned int cnt) __attribute__ ((optimize("-O3"))); // Secure fixed optimization, the "Ofast" will unfold the inner loop and spoil the timing
void sendNeoPixelData(unsigned int cnt) {
  bitClear(PORTB, PB5);       // Previous state of port is unknown
  delayMicroseconds(50);                      // Time for pixels to prepare for new data, is only needed if port was high
  noInterrupts();
  unsigned int idx = 0;
  byte n = nSerialBuffer;
  for (; cnt > 0; cnt--) {            // The timing between bytes is way above specifications
    byte b = serialBuffer[idx];
    idx = (idx + 1) % n;
    for (byte i = 0; i < (byte) 8; i++) {
      if (b & (byte) 0x80) {                // Need 800ns high and 450ns low. The low time is handled by the preparation for next bit/byte
        bitSet( PORTB, PB5);
        asm volatile(".rept 8\n\tNOP\n\t.endr":::);
        bitClear( PORTB, PB5);
      } else {                            // Need 400ns (Some chips are nominal 300ns) high and 850ns low. Including some extra low time here.
        bitSet( PORTB, PB5);
        asm volatile(".rept 2\n\tNOP\n\t.endr":::); // This is the only critical timing
        bitClear( PORTB, PB5);
        asm volatile(".rept 7\n\tNOP\n\t.endr":::);
      }
      asm volatile(".rept 1\n\tNOP\n\t.endr":::);
      b <<= (byte) 1;
    }
  }
  interrupts();
  // Time for neopixels to latch data, they can do that before next call
}


// For APA102 SPI clocked led pixels, Data PB5, clock PC6
// This send a n byte prefix with 0 bytes
void sendPixelSPIPrefix(byte cnt) {
  bitClear(PORTB, PB5);
  bitSet( PORTC, PC6);
  for (; cnt > 0; cnt--) {
    for (byte i = 0; i < (byte) 8; i++) {
      asm volatile(".rept 3\n\tNOP\n\t.endr":::);
      bitClear( PORTC, PC6);
      asm volatile(".rept 6\n\tNOP\n\t.endr":::);
      bitSet( PORTC, PC6);
    }
  }
}

void sendPixelSPIPostfix(byte cnt) {
  bitSet(PORTB, PB5);
  bitSet( PORTC, PC6);
  for (; cnt > 0; cnt--) {
    for (byte i = 0; i < (byte) 8; i++) {
      asm volatile(".rept 3\n\tNOP\n\t.endr":::);
      bitClear( PORTC, PC6);
      asm volatile(".rept 6\n\tNOP\n\t.endr":::);
      bitSet( PORTC, PC6);
    }
    bitClear(PORTB, PB5); // Datasheet says to use all 1's, but this works and the led after last will not turn on.
  }
  asm volatile(".rept 6\n\tNOP\n\t.endr":::);
  bitClear( PORTC, PC6);
}


// For APA102 SPI clocked led pixels, Data PB5, clock PC6
void sendPixelSPIData(unsigned int cnt) {
  bitSet( PORTB, PB5);
  bitSet( PORTC, PC6);
  unsigned int idx = 0;
  byte n = nSerialBuffer;
  for (; cnt > 0; cnt--) {
    byte b = serialBuffer[idx];
    idx = (idx + 1) % n;
    for (byte i = 0; i < (byte) 8; i++) {
      if ((b & (byte) 0x80)) bitSet(PORTB, PB5); else bitClear(PORTB, PB5);
      asm volatile(".rept 6\n\tNOP\n\t.endr":::);
      bitClear( PORTC, PC6);
      asm volatile(".rept 6\n\tNOP\n\t.endr":::);
      bitSet( PORTC, PC6);
      b <<= (byte) 1;
    }
  }
}



//----------------------------------------------------------------------------------------------------
// Refresh led output, it is just a copy of mos outputs, this routine is called when the Arduino is idle.
void setLeds() {
  // Pulse stretcher
  if (ledPulse[0] > 0) ledPulse[0]--;
  if (ledPulse[1] > 0) ledPulse[1]--;
  if (ledPulse[2] > 0) ledPulse[2]--;
  if (ledPulse[3] > 0) ledPulse[3]--;
  if (PINB & 0x20) ledPulse[0] = 100;
  if (PINB & 0x40) ledPulse[1] = 100;
  if (PINC & 0x40) ledPulse[2] = 100;
  if (inProgram) ledPulse[3] = 100;

  // In PWM modes the leds have trouble picking up short pulses, switching to constant brightness at higher frequencies helps.
  // But very low duty cycle at low frequencies do not work reliable (With the current code the led will track the output at moderate duty cycle levels).
  if (fastpwm && mode == MODE_PWM) {
    if (pwmDutyCycle1 > 0) ledPulse[0] = 100;
    if (pwmDutyCycle2 > 0) ledPulse[1] = 100;
    if (pwmDutyCycle3 > 0) ledPulse[2] = 100;
  } else if (fastpwm && mode == MODE_PPPWM) {
    if (pwmDutyCycle1 > 0) {
      ledPulse[0] = 100;
      ledPulse[1] = 100;
    }
  }

  byte pd = PORTD & ~(1 << PD1 | 1 << PD0 | 1 << PD4 | 1 << PD7);
  pd |= (ledPulse[0] != 0) ? 1 << PD1 : 0;
  pd |= (ledPulse[1] != 0) ? 1 << PD0 : 0;
  pd |= (ledPulse[2] != 0) ? 1 << PD4 : 0;
  pd |= (ledPulse[3] != 0) ? 1 << PD7 : 0;
  PORTD = pd;
}

//----------------------------------------------------------------------------------------------------
// Configure and start relay output mode
void setRelays() {
  if (mode != MODE_RELAY) clearOutputs();
  mode = MODE_RELAY;
  // Try to change all outputs simultaneous
  byte pb = PORTB & ~(1 << PB5 | 1 << PB6 | 1 << PB3);
  if (mosStatus & 1) pb |= (1 << PB5);
  if (mosStatus & 2) pb |= (1 << PB6);
  if (mosStatus & 8) pb |= (1 << PB3);
  noInterrupts();
  PORTC = (PORTC & ~(1 << PC6)) | ((mosStatus & 4) ? 1 << PC6 : 0);
  PORTB = pb;
  interrupts();
}
//----------------------------------------------------------------------------------------------------
// Calculate timers and divisor for Timer1 & Timer3 for all modes with time specification.
void calcPulseTimers(TimerSettings * ts, unsigned long t1, unsigned long t2, unsigned long t3) {
  if (t1 > 410000L) {
    t1 = 410000L;
  }
  if (t2 > 410000L) {
    t2 = 410000L;
  }
  if (t3 > 410000L) {
    t3 = 410000L;
  }

  long pulseTime = max(t1, t2);

  if (pulseTime < 1 * 65535 / 160) { // 16MHz is used as 160 to compensate for 0.01ms resolution
    ts->maskb1 = (1 << CS10); // CLK/1
    ts->fac1 = 1;
  } else if (pulseTime < 8L * 65535 / 160) {
    ts->maskb1 = (1 << CS11); // CLK/8
    ts->fac1 = 8;
  } else if (pulseTime < 64L * 65535 / 160) {
    ts->maskb1 = (1 << CS11) | (1 << CS10); // CLK/64
    ts->fac1 = 64;
  } else if (pulseTime < 256L * 65535 / 160) {
    ts->maskb1 = (1 << CS12); // CLK/256
    ts->fac1 = 256;
  } else {
    ts->maskb1 = (1 << CS12) | (1 << CS10); // CLK/1024
    ts->fac1 = 1024;
  }
  ts->o1 = (unsigned int) (t1 * 160L / ts->fac1 ) - 1;
  ts->o2 = (unsigned int) (t2 * 160L / ts->fac1 ) - 1;
  /*
    debug->print("  t1:");
    debug->print(t1);
    debug->print("  o1:");
    debug->print(ts->o1);
    debug->print("  fac:");
    debug->print(ts->fac1);
    debug->print("  maskb:");
    debug->print(ts->maskb1);
    debug->println();
  */


  pulseTime = t3;
  if (pulseTime < 1 * 65535 / 160) {
    ts->maskb3 = (1 << CS30); // CLK/8
    ts->fac3 = 1;
  } else if (pulseTime < 8 * 65535 / 160) {
    ts->maskb3 = (1 << CS31); // CLK/8
    ts->fac3 = 8;
  } else if (pulseTime < 64 * 65535 / 160) {
    ts->maskb3 = (1 << CS31) | (1 << CS30); // CLK/64
    ts->fac3 = 64;
  } else if (pulseTime < 256L * 65535 / 160) {
    ts->maskb3 = (1 << CS32); // CLK/256
    ts->fac3 = 256;
  } else {
    ts->maskb3 = (1 << CS32) | (1 << CS30); // CLK/1024
    ts->fac3 = 1024;
  }
  ts->o3 = (unsigned int) (t3 * 160L / ts->fac3) - 1;
}

//----------------------------------------------------------------------------------------------------
// Configure and start sequency output mode.
void setSequence() {
  clearOutputs();
  mode = MODE_SEQ;
  mosStatus = 0;
  TimerSettings ts;
  calcPulseTimers(&ts, pulseT1, pulseT2, pulseT3);
  resolution = max(ts.o1, max(ts.o2, ts.o3));
  /*
    debug->print("Ready 1:");
    debug->print(ts.o1);
    debug->print("  2:");
    debug->print(ts.o2);
    debug->print("  3:");
    debug->print(ts.o3);
    debug->print("  mask:");
    debug->print(ts.maskb1, 2);
    debug->print(' ');
    debug->print(ts.maskb3, 2);
    debug->println();
  */
  noInterrupts();           // disable all interrupts
  OCR1A = max(2, ts.o1);
  OCR1B = max(2, ts.o2);
  OCR3A = max(2, ts.o3);
  TCNT1 = 0;
  TCNT3 = 0;
  TCCR1A = (1 << COM1A1) | (1 << COM1B1);   // Force outputs low on match
  TCCR3A = (1 << COM3A1);
  TCCR1C = (1 << FOC1A) | (1 << FOC1B); // Reset all outputs
  TCCR3C = (1 << FOC3A);

  TCCR1A = (1 << COM1A1) | (1 << COM1B1) | (1 << COM1A0) | (1 << COM1B0); // Set outputs on match
  TCCR3A = (1 << COM3A1) | (1 << COM3A0);

  TCCR1B = ts.maskb1;    // Start timers
  TCCR3B = ts.maskb3;
  interrupts();             // enable all interrupts
}

//----------------------------------------------------------------------------------------------------
void setSlowSequence() {
  if (mode != MODE_SLOW_SEQ) clearOutputs();
  mode = MODE_SLOW_SEQ;
  mosStatus = 0;
  digitalWrite(mos1Pin, LOW);
  digitalWrite(mos2Pin, LOW);
  digitalWrite(mos3Pin, LOW);
  unsigned next = min(pulseT1, min(pulseT2, pulseT3));  // First event

  slowTimer = next;
  nextSlowEvent = millis() + next / 100;
}

//----------------------------------------------------------------------------------------------------
void updateSlowSequence() {
  unsigned long next = 0xffffffffUL;
  if (slowTimer >= pulseT1) {
    digitalWrite(mos1Pin, HIGH);
  } else {
    next = min(next, pulseT1);
  }
  if (slowTimer >= pulseT2) {
    digitalWrite(mos2Pin, HIGH);
  } else {
    next = min(next, pulseT2);
  }
  if (slowTimer >= pulseT3) {
    digitalWrite(mos3Pin, HIGH);
  } else {
    next = min(next, pulseT3);
  }

  long dt = next - slowTimer; // Time to next event
  if (dt > 0) {
    slowTimer += dt;
    nextSlowEvent +=  dt / 100;
  } else {
    nextSlowEvent = 0xffffffffUL;   // No more events, but outputs stays high
  }
}


//----------------------------------------------------------------------------------------------------
// Configure and start pulse all mode
void setPulseAll() {
  clearOutputs();
  mode = MODE_PULSEALL;
  mosStatus = 0;
  TimerSettings ts;
  calcPulseTimers(&ts, pulseT1, pulseT2, pulseT3);

  byte reg1a = 0;
  byte reg3a = 0;
  if (ts.o1 > 0) reg1a |= (1 << COM1A1) | (1 << COM1A0);
  if (ts.o2 > 0) reg1a |= (1 << COM1B1) | (1 << COM1B0);
  if (ts.o3 > 0) reg3a |= (1 << COM3A1) | (1 << COM3A0);

  resolution = max(ts.o1, max(ts.o2, ts.o3));

  /*
    debug->print("T1:");
    debug->print(ts.o1);
    debug->print("  T2:");
    debug->print(ts.o2);
    debug->print("  T3:");
    debug->print(ts.o3);

    debug->print("  maskb1:");
    debug->print(ts.maskb1,2);
    debug->print("  maskb3:");
    debug->print(ts.maskb3,2);
    debug->println();
  */

  noInterrupts();           // disable all interrupts
  TCNT1 = 0;
  TCNT3 = 0;
  OCR1A = max(2, ts.o1);
  OCR1B = max(2, ts.o2);
  OCR3A = max(2, ts.o3);
  TCCR1A = reg1a;
  TCCR3A = reg3a;
  TCCR1B = ts.maskb1;                    // Set outputs
  TCCR3B = ts.maskb3;
  TCCR1C = (1 << FOC1A) | (1 << FOC1B);
  TCCR3C = (1 << FOC3A);
  TCCR1A = (1 << COM1A1) | (1 << COM1B1); // Clear output on match
  TCCR3A = (1 << COM3A1);
  interrupts();             // enable all interrupts
}


//----------------------------------------------------------------------------------------------------
void setSlowPulseAll() {
  if (mode != MODE_SLOW_PULSEALL) clearOutputs();
  mode = MODE_SLOW_PULSEALL;
  mosStatus = 0;
  digitalWrite(mos1Pin, HIGH);
  digitalWrite(mos2Pin, HIGH);
  digitalWrite(mos3Pin, HIGH);
  unsigned next = min(pulseT1, min(pulseT2, pulseT3));  // First event

  slowTimer = next;
  nextSlowEvent = millis() + next / 100;
}

//----------------------------------------------------------------------------------------------------
void updateSlowPulseAll() {
  unsigned long next = 0xffffffffUL;
  if (slowTimer >= pulseT1) {
    digitalWrite(mos1Pin, LOW);
  } else {
    next = min(next, pulseT1);
  }
  if (slowTimer >= pulseT2) {
    digitalWrite(mos2Pin, LOW);
  } else {
    next = min(next, pulseT2);
  }
  if (slowTimer >= pulseT3) {
    digitalWrite(mos3Pin, LOW);
  } else {
    next = min(next, pulseT3);
  }

  long dt = next - slowTimer; // Time to next event
  if (dt > 0) {
    slowTimer += dt;
    nextSlowEvent +=  dt / 100;
  } else {
    nextSlowEvent = 0xffffffffUL;   // No more events
  }
}



//----------------------------------------------------------------------------------------------------
// Configure and start pattern mode
void setPattern() {
  clearOutputs();
  mode = MODE_PATTERN;
  mosStatus = 0;
  patternIdx = 0;
  TimerSettings ts;
  calcPulseTimers(&ts, pulseT1, pulseT2, pulseT3);

  // Use timer 1 as interrupt source

  byte tccr1aMask = (1 << WGM11); // fast mode, TOP is ICR1
  byte tccrbMask = (1 << WGM13) | (1 << WGM12); // fast mode

  tccrbMask |= ts.maskb1;
  resolution = ts.o1;

  /*
    debug->print("cnt:");
    debug->print(ts.o1);
    debug->print("  amask:");
    debug->print(tccr1aMask, 2);
    debug->print("  bmask:");
    debug->print(tccrbMask, 2);
    debug->print("  len:");
    debug->print(npattern);
    debug->println();
  */
  noInterrupts();           // disable all interrupts
  TCNT1 = 0;
  TIFR1 = 0;              // Clear any pending interrupts
  ICR1 = ts.o1;
  TCCR1A = tccr1aMask;
  TCCR1B = tccrbMask;
  TIMSK1 = (1 << TOIE1);   // enable timer overflow interrupt
  interrupts();             // enable all interrupts

}

//----------------------------------------------------------------------------------------------------
void setSlowPattern() {
  if (mode != MODE_SLOW_PATTERN) clearOutputs();
  mode = MODE_SLOW_PATTERN;
  mosStatus = 0;
  patternIdx = 0;
  nextSlowEvent = millis(); // Trig first event immediately
  slowUpdateInterval = pulseT1 / 100;
}

//----------------------------------------------------------------------------------------------------
void updateSlowPattern() {
  nextSlowEvent += slowUpdateInterval;
  byte p = pattern[patternIdx];
  PORTB = (PORTB & ~(0x64)) | (p & 1 ? 0x20 : 0) | (p & 2 ? 0x40 : 0) | (p & 8 ? 0x04 : 0);
  PORTC = (PORTC & ~(0x40)) | (p & 4 ? 0x40 : 0);
  if (repeatPattern) patternIdx = (patternIdx + 1) % npattern;
  else if (patternIdx < npattern - 1) {
    patternIdx++;
  }
}


//----------------------------------------------------------------------------------------------------
// Configure and start sing pulse mode.
void setPulse() {
  clearOutputs();
  mode = MODE_PULSE;
  mosStatus = 0;
  TimerSettings ts;
  calcPulseTimers(&ts, pulseTime, pulseTime, pulseTime);
  byte tccr1aResetMask = 0;
  byte tccr1aSetMask = 0;
  byte foc1Mask = 0;
  byte tccr3aResetMask = 0;
  byte tccr3aSetMask = 0;
  byte foc3Mask = 0;

  switch (pulseNo) {
    case 0:
      tccr1aResetMask |= (1 << COM1A1) | (1 << COM1B1);
      tccr1aSetMask |= (1 << COM1A1) | (1 << COM1B1) | (1 << COM1A0) | (1 << COM1B0);
      foc1Mask = (1 << FOC1A) | (1 << FOC1B);
      tccr3aResetMask |= (1 << COM3A1);
      tccr3aSetMask |= (1 << COM3A1) | (1 << COM3A0);
      foc3Mask = (1 << FOC3A);
      noInterrupts();           // disable all interrupts
      OCR1A = ts.o1;
      OCR1B = ts.o2;
      OCR3A = ts.o3;
      interrupts();             // enable all interrupts
      break;
    case 1:
      tccr1aResetMask |= (1 << COM1A1);
      tccr1aSetMask |= (1 << COM1A1) | (1 << COM1A0);
      foc1Mask = (1 << FOC1A);
      noInterrupts();           // disable all interrupts
      OCR1A = ts.o1;
      interrupts();             // enable all interrupts
      break;
    case 2:
      tccr1aResetMask |= (1 << COM1B1);
      tccr1aSetMask |= (1 << COM1B1) | (1 << COM1B0);
      foc1Mask = (1 << FOC1B);
      noInterrupts();           // disable all interrupts
      OCR1B = ts.o2;
      interrupts();             // enable all interrupts
      break;
    case 3:
      tccr3aResetMask |= (1 << COM3A1);
      tccr3aSetMask |= (1 << COM3A1) | (1 << COM3A0);
      foc3Mask = (1 << FOC3A);
      noInterrupts();           // disable all interrupts
      OCR3A = ts.o3;
      interrupts();             // enable all interrupts
      break;
  }

  resolution = max(ts.o1, max(ts.o2, ts.o3));

  noInterrupts();           // disable all interrupts
  TCNT1 = 0;
  TCNT3 = 0;
  TCCR1A = tccr1aSetMask;
  TCCR1B = ts.maskb1;
  TCCR1C = foc1Mask;
  TCCR1A = tccr1aResetMask;

  TCCR3A = tccr3aSetMask;
  TCCR3B = ts.maskb3;
  TCCR3C = foc3Mask;
  TCCR3A = tccr3aResetMask;
  interrupts();             // enable all interrupts
}


//----------------------------------------------------------------------------------------------------
void setSlowPulse() {
  if (mode != MODE_SLOW_PULSE) clearOutputs();
  mode = MODE_SLOW_PULSE;
  mosStatus = 0;
  switch (pulseNo) {
    case 0 :
      digitalWrite(mos1Pin, HIGH);
      digitalWrite(mos2Pin, HIGH);
      digitalWrite(mos3Pin, HIGH);
      break;
    case 1 :
      digitalWrite(mos1Pin, HIGH);
      break;
    case 2 :
      digitalWrite(mos2Pin, HIGH);
      break;
    case 3 :
      digitalWrite(mos3Pin, HIGH);
      break;
  }
  nextSlowEvent = millis() + pulseTime / 100;
}

//----------------------------------------------------------------------------------------------------
void updateSlowPulse() {
  setRelays();
}


//----------------------------------------------------------------------------------------------------
void setSlowPulses() {
  clearOutputs();
  mode = MODE_SLOW_PULSES;
  mosStatus = 0;
  digitalWrite(mos1Pin, HIGH);
  nextSlowEvent = millis() + pulseTime / 100;
}


//----------------------------------------------------------------------------------------------------
void updateSlowPulses() {
  boolean next = !digitalRead(mos1Pin);
  if (next) {
    pulseCount--;
    if (pulseCount == 0) {
      setRelays();
      return;
    }
  }
  digitalWrite(mos1Pin, next);
  nextSlowEvent = millis() + (next ? pulseTime : pulseTimeOff) / 100;
}



//----------------------------------------------------------------------------------------------------
// Configure and start pwm push pull mode PPPWM
void setPushPullDrive() {
  if (mode != MODE_PPPWM) clearOutputs();
  mode = MODE_PPPWM;
  mosStatus = 0;
  long maxPeriod;
  // Switch to timer 3 as intr source
  TCCR3A = 0;
  TCCR3B = (1 << CS30);    // 1 prescaler 250Khz or 4ms intr
  TIMSK3 = (1 << TOIE3);   // enable timer overflow interrupt
  TIMSK4 = 0;             // No interrupt from Timer4

  if ((PLLCSR & (1 << PLLE)) == 0) { // Check if PLL is disabled, it is used for USB
    PLLCSR  = (1 << PINDIV);  // set 16MHz clock system clock.
    PLLFRQ |= (1 << PDIV3) | (1 << PDIV1); // 96MHz PLL frequency.
    PLLCSR |= (1 << PLLE);    // enable PLL.
    while ((PLLCSR & (1 << PLOCK)) == 0); // Wait for PLL lock.
  }

  PLLFRQ |= (1 << PLLTM1) | (1 << PLLTM0);  // Divide PLL by 2, i.e. boost internal timer 4 clock to 48MHz

  byte reg4b = 0;

  reg4b = (1 << DTPS41) | (1 << DTPS40);    // Dead time /8
  int d = deadtime / 16;

  /* This code do not work, the divisor is ignored
    int d = 0;
    if (deadtime > 120) {
      reg4b = (1 << DTPS41) | (1 << DTPS40);    // Dead time /8
      d = deadtime / 16;
    } else  if (deadtime > 60) {
      reg4b = (1 << DTPS41);    // Dead time /4
      d = deadtime / 8;
    } else  if (deadtime > 30) {
      reg4b = (1 << DTPS40);    // Dead time /2
      d = deadtime / 4;
    } else {
      reg4b = 0;    // Dead time /1
      d = deadtime / 2;
    }*/
  if (d > 15) d = 15;
  DT4 = (d << 4) | d;

  int top = 0;
  int factor = 1;
  int code = 1;

  resolution = top;

  if (timeFlag & FREQ_TIME) {
    if (pwmFrequency > 69000L) {
      pwmFrequency = 69000L;
    } else if (pwmFrequency <= 1) {
      pwmFrequency = 1;
    }
    maxPeriod = pwmFrequency;
    factor = 1;
    code = 1;
    while (pwmFrequency > 1023L * factor * 2 / 160 / 3 && code != 0x0f) {
      factor *= 2;
      code++;
    }
    fastpwm = pwmFrequency < 10000;
    top = pwmFrequency * 160L * 3 / 2 / factor;
  } else {
    // Limit pwm frequencies to supported range
    if (pwmFrequency > 1000000L) {
      pwmFrequency = 1000000L;
    } else if (pwmFrequency <= 15) {
      pwmFrequency = 15;
    }
    factor = 1;
    code = 1;
    while (pwmFrequency <= 160000000L * 3 / 1023 / 2 / factor && code != 0x0f) {
      factor *= 2;
      code++;
    }
    top = 160000000L * 3 / 2 / factor / pwmFrequency;
    fastpwm = pwmFrequency > 100;
    maxPeriod = 1000000 / pwmFrequency;
  }

  int dc;
  if (timeFlag & PWM1_TIME) {
    if (pwmDutyCycle1 > maxPeriod) pwmDutyCycle1 = maxPeriod;
    dc = (long) pwmDutyCycle1 * 160L * 3 / 2 / factor;
    if (dc > top) dc = top;
  } else {
    dc = (long) pwmDutyCycle1 * top / 1000L;
  }
  dc = top - dc;
  reg4b |= code;

  /*
      debug->print("Factor:");
      debug->print(factor);
      debug->print("  Top:");
      debug->print(top);
      debug->print("  dc:");
      debug->print(dc);

      debug->print("  dtm:");
      debug->print(reg4b, 2);
      debug->print("  dt:");
      debug->print(d);
      debug->print((pwmLastFreq != pwmFrequency)?"  FRQ":"");
      debug->println();
  */

  resolution = top;
  noInterrupts();           // disable all interrupts
  if (pwmLastFreq != pwmFrequency) {
    TCCR4A = (1 << COM4B0) | (1 << PWM4B);  // Enable !PB5 & PB6
    TCCR4D = (1 << WGM41) | (1 << WGM40);   // PWM6 dual slope with OCR4C as TOP
    TCCR4B = reg4b;

    TC4H = (top >> 8);        // Set top
    OCR4C = (top & 0xff);
  }

  TC4H = (dc >> 8);        // Set duty cycle
  OCR4B = dc & 0xff;
  interrupts();           // enable all interrupts
  pwmLastFreq = pwmFrequency;
}

//----------------------------------------------------------------------------------------------------
// Configure and start PWM mode
void setPWM() {
  if (mode != MODE_PWM) clearOutputs();
  mode = MODE_PWM;
  mosStatus = 0;
  long maxPeriod;

  byte tccr1aMask = (1 << WGM11); // fast mode 14, TOP is ICR1
  byte tccr3aMask = (1 << WGM11); // fast mode 14, TOP is ICR1
  byte tccrbMask = (1 << WGM13) | (1 << WGM12); // fast mode 14
  // A duty cycle of 0 will disable pwm on that output to prevent a 1 cycle spike.
  if (pwmDutyCycle1 > 0) {
    tccr1aMask |= (1 << COM1A1);
  }
  if (pwmDutyCycle2 > 0) {
    tccr1aMask |= (1 << COM1B1);
  }
  if (pwmDutyCycle3 > 0) {
    tccr3aMask |= (1 << COM3A1);
  }

  int fac = 1;
  unsigned int tcrtop = 0;
  if (timeFlag & FREQ_TIME) {
    // Limit pwm periods to supported range
    if (pwmFrequency > 400000L) {
      pwmFrequency = 400000L;
    } else if (pwmFrequency <= 0) {
      pwmFrequency = 1;
    }
    maxPeriod = pwmFrequency;
    fastpwm = pwmFrequency < 10000;
    TimerSettings ts;
    calcPulseTimers(&ts, pwmFrequency, pwmFrequency, pwmFrequency);
    tccrbMask |= ts.maskb1;
    tcrtop = ts.o1;
    fac = ts.fac1;
  } else {
    // Limit pwm frequencies to supported range
    if (pwmFrequency > 1000000L) {
      pwmFrequency = 1000000L;
    } else if (pwmFrequency <= 3) {
      pwmFrequency = 3;
    }
    maxPeriod = 1000000 / pwmFrequency;
    if (pwmFrequency > 160000000 / 65535) {
      tccrbMask |= (1 << CS10); // CLK/1
      fac = 1;
    } else if (pwmFrequency > 160000000 / 65535 / 8) {
      tccrbMask |= (1 << CS11); // CLK/8
      fac = 8;
    } else if (pwmFrequency > 160000000 / 65535 / 64) {
      tccrbMask |= (1 << CS11) | (1 << CS10); // CLK/64
      fac = 64;
    } else if (pwmFrequency > 160000000 / 65535 / 256) {
      tccrbMask |= (1 << CS12); // CLK/256
      fac = 256;
    } else {
      tccrbMask |= (1 << CS12) | (1 << CS10); // CLK/1024
      fac = 1024;
    }
    fastpwm = pwmFrequency > 100;
    tcrtop = (unsigned int) (160000000L / fac / pwmFrequency);
  }
  unsigned int dc1 = 0;
  unsigned int dc2 = 0;
  unsigned int dc3 = 0;

  if (timeFlag & PWM1_TIME) {
    if (pwmDutyCycle1 > maxPeriod) pwmDutyCycle1 = maxPeriod;
    dc1 = (long) pwmDutyCycle1 * 160 / fac;
    if (dc1 > tcrtop) dc1 = tcrtop;
  } else {
    dc1 = (long) pwmDutyCycle1 * tcrtop / 1000L;
  }

  if (timeFlag & PWM2_TIME) {
    if (pwmDutyCycle2 > maxPeriod) pwmDutyCycle2 = maxPeriod;
    dc2 = (long) pwmDutyCycle2 * 160 / fac;
    if (dc2 > tcrtop) dc2 = tcrtop;
  } else {
    dc2 = (long) pwmDutyCycle2 * tcrtop / 1000L;
  }

  if (timeFlag & PWM3_TIME) {
    if (pwmDutyCycle3 > maxPeriod) pwmDutyCycle3 = maxPeriod;
    dc3 = (long) pwmDutyCycle3 * 160 / fac;
    if (dc3 > tcrtop) dc3 = tcrtop;
  } else {
    dc3 = (long) pwmDutyCycle3 * tcrtop / 1000L;
  }

  /*
    debug->print("top: ");
    debug->print(tcrtop);
    debug->print("  dc1:");
    debug->print(dc1);
    debug->print("  dc2:");
    debug->print(dc2);
    debug->print("  dc3:");
    debug->print(dc3);
    debug->print("  maskb:");
    debug->print(tccrbMask, 2);
    debug->print((pwmLastFreq != pwmFrequency)?"  FRQ":"");
    debug->println();
  */
  resolution = tcrtop;

  noInterrupts();           // disable all interrupts
  if (pwmLastFreq != pwmFrequency) {
    ICR1 = tcrtop - 1; // Define top
    ICR3 = tcrtop - 1;
    TCNT1 = 0;      // Reset counters
    TCNT3 = 0;
  }
  OCR1A = dc1;    // Set new PWM comperators
  OCR1B = dc2;
  OCR3A = dc3;

  TCCR1A = tccr1aMask;  // Start timers with new settings, this may change divider
  TCCR1B = tccrbMask;
  TCCR3A = tccr3aMask;
  TCCR3B = tccrbMask;
  interrupts();             // enable all interrupts
  pwmLastFreq = pwmFrequency;

}

//----------------------------------------------------------------------------------------------------

void setPulses() {
  clearOutputs();
  mode = MODE_PULSES;
  TimerSettings ts;
  calcPulseTimers(&ts, pulseTime + pulseTimeOff, pulseTime, 0);
  resolution = ts.o2;
  ts.maskb1 |= (1 << WGM12) | (1 << WGM13); // Mode 14 fast PWM
  noInterrupts();
  TCNT1 = ts.o1 - 5;
  TIMSK1 = (1 << OCIE1A); // Interrupt on compare match A
  ICR1 = ts.o1;
  OCR1A = ts.o2;
  TCCR1A = (1 << WGM11) | (1 << COM1A1); // Clear on compare match A
  TCCR1B = ts.maskb1;   // Start
  interrupts();
}

//----------------------------------------------------------------------------------------------------
// Define serial output
void setSerial(boolean repeat) {
  byte oldMode = mode;
  if (mode != MODE_SERIAL) clearOutputs();
  mode = MODE_SERIAL;
  if (oldMode != MODE_SERIAL) {
    digitalWrite(mos2Pin, LOW);
    pinMode(mos1Pin, INPUT);
    pinMode(mos1SerialPin, OUTPUT);   // This forces the state to low
    digitalWrite(mos1SerialPin, HIGH);
  }

  UCSR1B = 0;             // Disable transmitter and intr
  byte maska = (1 < U2X1);    // Double transmission speed, this will get removed at very low speeds, but because I only transmit there is no disadvantage with it
  byte maskb = (1 << UDRIE1) | (1 << TXEN1) | (1 << TXCIE1); // Enable transmitter & transmitter buffer empty interrupt and transmit complete interrupt.
  //byte maskb = (1 << TXEN1); // Enable transmitter
  byte maskc = 0;
  byte bits = (serialFormat & 0x7) + 5;
  repeatSerialBuffer = repeat;

  switch ((serialFormat >> 5) & 0x07) {
    case 0: maskc |= 0; break; // none
    case 1: maskc |= 0; bits++; break; // Space
    case 2: maskc |= 0; bits++; break; // Mark
    case 3: maskc |= (1 << UPM11) | (1 << UPM10); break; // Odd
    case 4: maskc = (1 << UPM11); break; // Odd
    case 7: maskc |= 0; bits++; break; // Command
  }
  if (serialFormat & 0x10) maskc |= (1 << USBS1);
  switch (bits) {
    case 5: maskc |= 0; break;
    case 6: maskc |= (1 << UCSZ10); break;
    case 7: maskc |= (1 << UCSZ11); break;
    case 8: maskc |= (1 << UCSZ11) | (1 << UCSZ10); break;
    case 9: maskc |= (1 << UCSZ11) | (1 << UCSZ10); maskb |= (1 << UCSZ12); break;
  }

  serialBufferParityMask = 0;
  if (((serialFormat >> 5) & 0x07) == 2) {  // Mark parity need to add a 1 to the data
    serialBufferParityMask = 1 << (bits - 1);
    if (serialBufferParityMask == 0) serialBufferParityMask = 0x01;
  }
  if (((serialFormat >> 5) & 0x07) == 7) {  // Mark parity need to add a 1 to the first data byte
    serialBufferParityMask |= 0x02;
    if (bits == 9) serialBufferParityMask |= 0x01;
    else serialBufferParityMask |= 1 << (bits - 1);
  }
  /*
    debug->print("maskc: ");
    debug->print(maskc, 2);
    debug->print("  Parity mask: ");
    debug->print(serialBufferParityMask, 16);
    debug->println();
  */
  byte f = (byte)16;
  long bdiv = ((16000000L / 16) + baudrate / 2) / baudrate  - 1; // the +baudrate/2 is for rounding purpose
  if (bdiv > 4095) {
    bdiv = 16000000L / baudrate / 32 - 1;
    if (bdiv > 4095) bdiv = 4095; // Maximum divider
    maska &= ~(1 < U2X1);
    f = (byte)32;
  } else if (bdiv < 0) bdiv = 0;
  baudrate = 16000000L / (bdiv + 1) / f; // Save the actual baudrate

  setLeds();
  delay((bits + 3) * 1000 / baudrate + 1); // One character time for serial line to stabilize on high, this is relevant both when breakin into a datastream or starting a new datastream
  UBRR1 = (int) bdiv;   // This register has 12 bits
  resolution = bdiv;

  serialBufferIdx = 0;
  noInterrupts();           // disable all interrupts
  UCSR1C = maskc;
  UCSR1B = maskb;
  UCSR1A = maska;
  interrupts();             // enable all interrupts
}
//----------------------------------------------------------------------------------------------------
// Define Pixel output
void setPixel() {
  if (nSerialBuffer == 0) return;
  if (mode != MODE_PIXEL) clearOutputs();
  mode = MODE_PIXEL;
  byte bytes;
  switch (pixelFormat & 0x07) {
    case 0: // RGB, GRB
    case 1:
      bytes = 3;
      if (pixelCount == 0) sendNeoPixelData(nSerialBuffer);
      else sendNeoPixelData(pixelCount * bytes);
      break;

    case 2: // RGBW, GRBW
    case 3:
      bytes = 4;
      if (pixelCount == 0) sendNeoPixelData(nSerialBuffer);
      else sendNeoPixelData(pixelCount * bytes);
      break;
    case 4: // APA102, SPI based
      bytes = 4;
      sendPixelSPIPrefix(bytes);   // Start frame
      if (pixelCount == 0) sendPixelSPIData(nSerialBuffer);
      else sendPixelSPIData(pixelCount * bytes);
      sendPixelSPIPostfix(bytes * (pixelCount / 60 + 1) + 1); // End frame
      break;
  }
}

//----------------------------------------------------------------------------------------------------
// Define SPI output
void setSPI() {
  if (nSerialBuffer == 0) return;
  byte oldMode = mode;
  if (mode != MODE_SPI) clearOutputs();
  mode = MODE_SPI;
  if (oldMode != MODE_SPI) {
    digitalWrite(mos2Pin, spiFormat & 0x08 ? HIGH : LOW);
    pinMode(mos3Pin, INPUT);
    pinMode(mos1Pin, INPUT);
    pinMode(mos1MOSIPin, OUTPUT);
    pinMode(mos3SCKPin, OUTPUT);
  }
  if (spiFormat & 0x08) PORTB |= (1 << PB6);           // Reset enable output
  else PORTB &= ~(1 << PB6);

  byte maskSpcr = (1 << SPIE) | (1 << SPE) | (spiFormat & 0x04 ? (1 << DORD) : 0) | (1 << MSTR) | (spiFormat & 0x01 ? (1 << CPOL) : 0) | (spiFormat & 0x02 ? (1 << CPHA) : 0);
  byte bdiv = 2;
  byte f = 0;
  while (12000000L / bdiv > spirate && bdiv < 128) bdiv <<= 1;
  spirate = 16000000L / bdiv;
  boolean spi2x = false;
  switch (bdiv) {
    case 2: spi2x = true;  break;
    case 4: spi2x = false; break;
    case 8: spi2x = true; f = (1 << SPR0); break;
    case 16: spi2x = false; f = (1 << SPR0); break;
    case 32: spi2x = true; f = (1 << SPR1); break;
    case 64: spi2x = false; f = (1 << SPR1); break;
    case 128: spi2x = true; f = (1 << SPR0) | (1 << SPR1); break;
  }
  resolution = 0;
  maskSpcr |= f;

  resolution = bdiv;
  serialBufferIdx = 0;
  noInterrupts();           // disable all interrupts
  SPSR = (spi2x ? 1 << SPI2X : 0);
  SPCR = maskSpcr;
  if (spiFormat & 0x08) PORTB &= ~(1 << PB6);           // Signal transmission starts now
  else PORTB |= (1 << PB6);
  SPDR = serialBuffer[serialBufferIdx++];
  interrupts();             // enable all interrupts
}

//----------------------------------------------------------------------------------------------------
// General time parser with support for some units
long parseTime(char *cmd) {
  byte l = strlen(cmd) - 1;
  float f = 100;
  if (!isDigit(*(cmd + l))) {
    switch (*(cmd + l)) {
      case 'u' : f = 0.1; break;  // usec
      case 's' : f = 100000.0; break; // seconds
      case 'm' : f = 6000000.0; break;  // minutes
      case 'h' : f = 360000000.0; break;  // hours
    }
    *(cmd + l) = 0;
  }
  return (unsigned long) (atof(cmd) * f);
}

/*
  long parseTime(char *cmd) {
  return (unsigned long) (atof(cmd) * 100);
  }
*/
//----------------------------------------------------------------------------------------------------
// Define sequency and select sequency mode.
void cmdSequence() {
  pulseT1 = 0;
  pulseT2 = 0;
  pulseT3 = 0;
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    pulseT1 = parseTime(cmd);
    cmd = strtok(NULL, paramDelimeters);
    if (cmd != NULL && *cmd != 0) {
      pulseT2 = parseTime(cmd);
      cmd = strtok(NULL, paramDelimeters);
      if (cmd != NULL && *cmd != 0) {
        pulseT3 = parseTime(cmd);
      }
    }
    if (max(pulseT1, max(pulseT2, pulseT3)) > 400000L) setSlowSequence();
    else setSequence();
  }
}


//----------------------------------------------------------------------------------------------------
// Restore program to ram, is called before any operation on the program
void restoreProgram() {
  int addr = PROGRAM_LOCATION;
  char *p = program;
  while (EEPROM.read(addr) != 0 && addr < MAX_PROGRAM_SIZE) {
    *p++ = EEPROM.read(addr);
    addr++;
  }
  *p = 0;
}


//----------------------------------------------------------------------------------------------------
// Called when program is chaged and will store program.
void saveProgram() {
  program[MAX_PROGRAM_SIZE - 1] = 0;
  char *p = program;
  int addr = PROGRAM_LOCATION;
  while (*p != 0) {
    if (*p == '|') *p = ';';
    if (EEPROM.read(addr) != *p) EEPROM.write(addr, *p);
    p++;
    addr++;
  }
  EEPROM.write(addr, 0);
}


//----------------------------------------------------------------------------------------------------
// Define stored program
void cmdProgram() {
  if (inProgram) return;
  char *cmds = strtok(NULL, "\r");
  strncpy(program, cmds, MAX_PROGRAM_SIZE);
  saveProgram();
}

//----------------------------------------------------------------------------------------------------
// Add statements to stored program
void cmdProgramAdd() {
  if (inProgram) return;
  restoreProgram();
  char *cmds = strtok(NULL, "\r");
  char *p = program;
  while (*p != 0) p++;
  *p++ = ';';
  strncpy(p, cmds, MAX_PROGRAM_SIZE - strlen(program) - 1);
  saveProgram();
}

//----------------------------------------------------------------------------------------------------
// List the stored program
void cmdProgramQ(Stream * stream) {
  restoreProgram();
  char buf[5];
  char *p = program;
  stream->print(itoa(strlen(p), buf, 10));
  stream->print(' ');
  while (*p != 0) {
    stream->print(*p);
    p++;
  }
  stream->print('\n');
}

//----------------------------------------------------------------------------------------------------
// Run a stored program
void cmdRun(Stream * stream) {
  loopProgram = false;
  if (inProgram) {
    loopProgram = true;
    return;
  }
  restoreProgram();
  inProgram = true;
  cmdProcessor.processCmd(program, stream);
  buttonPressed = false;  // Any button pressed while a program is running is used to stop program, not to reset output.
}

//----------------------------------------------------------------------------------------------------
// Define a pattern and select pattern mode, it can both be repeating or not.
void cmdPattern(boolean repeat) {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    pulseT1 = parseTime(cmd);
    cmd = strtok(NULL, paramDelimeters);
    if (cmd != NULL && *cmd != 0) {
      npattern = 0;
      while (*cmd != 0 && npattern < MAX_PATTERN_SIZE) {
        byte c = *cmd++;
        if (c >= 'a' && c <= 'f') c = c - 'a' + 10;
        else if (c >= 'A' && c <= 'F') c = c - 'A' + 10;
        else if (c >= '0' && c <= '9') c = c - '0';
        else c = 0;
        pattern[npattern] = c;
        npattern++;
      }
    }
  }
  repeatPattern = repeat;
  if (pulseT1 > 400000L) setSlowPattern();
  else setPattern();
}

//----------------------------------------------------------------------------------------------------
// Ask for current time and pattern
void cmdPatternQ(Stream * stream) {
  printTime(stream, pulseT1);
  stream->print(' ');
  for (int i = 0; i < npattern; i++) {
    char c = (char) pattern[i];
    c += (c < 10) ? '0' : ('A' - 10);
    stream->print(c);
  }
  stream->print('\n');
}

//----------------------------------------------------------------------------------------------------
// Ask for current resolution
void cmdResolutionQ(Stream * stream) {
  stream->print(resolution);
  stream->print('\n');
}

//----------------------------------------------------------------------------------------------------
// Specify pulse width for all outputs and select pulse all mode
void cmdPulseAll() {
  pulseT1 = 0;
  pulseT2 = 0;
  pulseT3 = 0;
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    pulseT1 = parseTime(cmd);
    cmd = strtok(NULL, paramDelimeters);
    if (cmd != NULL && *cmd != 0) {
      pulseT2 = parseTime(cmd);
      cmd = strtok(NULL, paramDelimeters);
      if (cmd != NULL && *cmd != 0) {
        pulseT3 = parseTime(cmd);
      }
    }
    if (max(pulseT1, max(pulseT2, pulseT3)) > 400000L) setSlowPulseAll();
    else setPulseAll();
  }
}

//----------------------------------------------------------------------------------------------------
// Specifiy output and pulse width and select single pulse mode.
void cmdPulse() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    pulseNo = atoi(cmd);
    if (pulseNo < 0 || pulseNo > 3) pulseNo = 0;
    cmd = strtok(NULL, paramDelimeters);
    if (cmd != NULL && *cmd != 0) {
      pulseTime = parseTime(cmd);
      if (pulseTime > 400000L) setSlowPulse();
      else setPulse();
    }
  }
}

//----------------------------------------------------------------------------------------------------
// Specifiy ontime offtime and number of pulses
void cmdPulses() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    pulseTime = parseTime(cmd);
    cmd = strtok(NULL, paramDelimeters);
    if (cmd != NULL && *cmd != 0) {
      pulseTimeOff = parseTime(cmd);
      cmd = strtok(NULL, paramDelimeters);
      if (cmd != NULL && *cmd != 0) {
        pulseCount = atoi(cmd);
        if (pulseTime + pulseTimeOff > 400000L) setSlowPulses();
        else setPulses();
      }
    }
  }
}

//----------------------------------------------------------------------------------------------------
// Ask for dead time
void cmdDeadTimeQ(Stream * stream) {
  stream->print(deadtime / 100.0, 2);
  stream->print("us\r\n");
}

//----------------------------------------------------------------------------------------------------
// Specify dead time, this is saved in EEPROM
void cmdDeadTime() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    deadtime = (int) (atof(cmd) * 100);
    if (deadtime > 250) deadtime = 250;
    EEPROM.write(DEADTIME_LOCATION, deadtime);
    if (mode == MODE_PPPWM) setPushPullDrive();
  }
}

//----------------------------------------------------------------------------------------------------
// Specify pwm duty cycle and select push pull pwm with dead time
void cmdPushPullDrive() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    timeFlag &= ~(PWM1_TIME | PWM2_TIME | PWM3_TIME);
    pwmDutyCycle1 = (int) (atof(cmd) * 10);
    setPushPullDrive();
  }
}

//----------------------------------------------------------------------------------------------------
// Specify pwm on time and select push pull pwm with dead time
void cmdPushPullDriveW() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    timeFlag |= PWM1_TIME;
    pwmDutyCycle1 = parseTime(cmd);
    setPushPullDrive();
  }
}

//----------------------------------------------------------------------------------------------------
// Specify pwm duty cycle and select pwm mode
void cmdPWM() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    int channel = atoi(cmd);
    cmd = strtok(NULL, paramDelimeters);
    if (cmd != NULL && *cmd != 0) {
      long dc = (long) (atof(cmd) * 10);
      if (dc < 0) dc = 0;
      else if (dc > 1000) dc = 1000;
      switch (channel) {
        case 0: pwmDutyCycle1 = dc; pwmDutyCycle2 = dc; pwmDutyCycle3 = dc; timeFlag &= ~(PWM1_TIME | PWM2_TIME | PWM3_TIME); break;
        case 1: pwmDutyCycle1 = dc; timeFlag &= ~(PWM1_TIME); break;
        case 2: pwmDutyCycle2 = dc; timeFlag &= ~(PWM2_TIME); break;
        case 3: pwmDutyCycle3 = dc; timeFlag &= ~(PWM3_TIME); break;
      }
    }
    setPWM();
  }
}

//----------------------------------------------------------------------------------------------------
// Specify pwm duty cycle in time instead of percent and select pwm mode
void cmdPWMW() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    int channel = atoi(cmd);
    cmd = strtok(NULL, paramDelimeters);
    if (cmd != NULL && *cmd != 0) {
      unsigned long dc = parseTime(cmd);
      if (dc > 400000L) dc = 400000L;
      switch (channel) {
        case 0: pwmDutyCycle1 = dc; pwmDutyCycle2 = dc; pwmDutyCycle3 = dc; timeFlag |= (PWM1_TIME | PWM2_TIME | PWM3_TIME); break;
        case 1: pwmDutyCycle1 = dc; timeFlag |= (PWM1_TIME); break;
        case 2: pwmDutyCycle2 = dc; timeFlag |= (PWM2_TIME); break;
        case 3: pwmDutyCycle3 = dc; timeFlag |= (PWM3_TIME); break;
      }
    }
    setPWM();
  }
}

//----------------------------------------------------------------------------------------------------
// Ask for actual pwm duty cycle
void cmdPWMQ(Stream * stream) {
  if (timeFlag & PWM1_TIME) {
    printTime(stream, pwmDutyCycle1);
  } else {
    stream->print(pwmDutyCycle1 / 10.0, 1);
    stream->print("%");
  }
  if (mode != MODE_PPPWM) {
    stream->print(' ');
    if (timeFlag & PWM2_TIME) {
      printTime(stream, pwmDutyCycle2);
      stream->print(' ');
    } else {
      stream->print(pwmDutyCycle2 / 10.0, 1);
      stream->print("% ");
    }
    if (timeFlag & PWM3_TIME) {
      printTime(stream, pwmDutyCycle3);
    } else {
      stream->print(pwmDutyCycle3 / 10.0, 1);
      stream->print("%");
    }
  }
  stream->print('\n');
}

//----------------------------------------------------------------------------------------------------
// Change pwm frequency
void cmdPWMFreq() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    byte l = strlen(cmd) - 1;
    float f = 10.0;
    if (*(cmd + l) == 'k') {
      *(cmd + l) = 0;
      f = 10000.0;
    }
    pwmFrequency = (long) (atof(cmd) * f);
    timeFlag &= ~(FREQ_TIME);
    pwmLastFreq = 0;    // Force reprogramming of divider and top value
    if (mode == MODE_PPPWM) setPushPullDrive();
    else if (mode == MODE_PWM) setPWM();
  }
}

//----------------------------------------------------------------------------------------------------
// Change pwm periode
void cmdPWMPeriod() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    pwmFrequency = parseTime(cmd);
    timeFlag |= (FREQ_TIME);
    pwmLastFreq = 0;    // Force reprogramming of divider and top value
    if (mode == MODE_PPPWM) setPushPullDrive();
    else if (mode == MODE_PWM) setPWM();
  }
}

//----------------------------------------------------------------------------------------------------
// Ask for actual pwm frequency or periode, will report in the last specified unit
void cmdPWMFreqQ(Stream * stream) {
  if (timeFlag & FREQ_TIME) {
    printTime(stream, pwmFrequency);
  } else {
    stream->print(pwmFrequency / 10.0, 1);
    stream->print("Hz");
  }
  stream->print('\n');

}

//----------------------------------------------------------------------------------------------------
// Use a PWM mode for servo control, that is 20ms periode time and 0.5ms to 1.5ms pulse with
void cmdServo() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    int channel = atoi(cmd);
    cmd = strtok(NULL, paramDelimeters);
    if (cmd != NULL && *cmd != 0) {
      int dc = atoi(cmd);
      if (dc < 0) dc = 0;
      else if (dc > 100) dc = 100;
      dc = dc + 100;  // Adjust to servo range
      timeFlag |= FREQ_TIME;
      pwmFrequency = 2000;
      switch (channel) {
        case 0: pwmDutyCycle1 = dc; pwmDutyCycle2 = dc; pwmDutyCycle3 = dc; timeFlag |= (PWM1_TIME | PWM2_TIME | PWM3_TIME); break;
        case 1: pwmDutyCycle1 = dc; timeFlag |= (PWM1_TIME); break;
        case 2: pwmDutyCycle2 = dc; timeFlag |= (PWM2_TIME); break;
        case 3: pwmDutyCycle3 = dc; timeFlag |= (PWM3_TIME); break;
      }
    }
    setPWM();
  }
}

//----------------------------------------------------------------------------------------------------
// Set relays from a bitmask and select output mode.
void cmdRelays() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    mosStatus = atoi(cmd) & 0x0f;
    setRelays();
  }
}

//----------------------------------------------------------------------------------------------------
// Turn outputs off and select output mode.
void cmdOff() {
  mosStatus = 0;
  setRelays();
}


//----------------------------------------------------------------------------------------------------
// This program delays the command queue for a specificed amount of milliseconds
void cmdWait() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    long v = parseTime(cmd) / 100;
    if (v > 100000) v = 100000;
    unsigned long endtime = millis() + v;
    // Wait until timeout or a key is pressed
    while (millis() < endtime && !(buttonPressed && (buttonStatus != 0))) {
      delay(1);
      setLeds();
      if (mode > MODE_SLOW_LIMIT) slowHandling();
    }
  }
}

//----------------------------------------------------------------------------------------------------
// Turn a specific output on
void cmdRelayOn() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    int v = atoi(cmd);
    if (v < 1 || v > 5) return;
    mosStatus |= 1 << (v - 1);
    setRelays();
  }
}

//----------------------------------------------------------------------------------------------------
// Turn a specific output off
void cmdRelayOff() {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    int v = atoi(cmd);
    if (v < 1 || v > 5) return;
    mosStatus &= ~(1 << (v - 1));
    setRelays();
  }
}

//----------------------------------------------------------------------------------------------------
// Return current output status for relay outputs, this ignores pwm and pulse modes.
void cmdRelaysQ(Stream * stream) {
  stream->print(mosStatus);
  stream->print('\n');
}

char *parseStringInput(char *p) {
  char deli = *p++;
  while (*p != 0 && *p != deli) {
    char c = *p++;
    if (c == '\\' && *p != 0) {
      switch (tolower(*p++)) {
        case '\\' : c = '\\'; break;
        case 'r' : c = '\r'; break;
        case 'n' : c = '\n'; break;
        case 'b' : c = '\b'; break;
        case 't' : c = '\t'; break;
        case 'f' : c = '\f'; break;
        case '"' : c = '"'; break;
        case '\'': c = '\''; break;
        case '0' : c = 0; break;
        default : c = '?';
      }
    }
    serialBuffer[nSerialBuffer++] = c;
  }
  return p;
}

// Logarithmic brightness level for the 0-9 index, without index the code will use 255
byte pixelLevel[] = {1, 2, 3, 7, 12, 20, 34, 56, 93, 154};
//----------------------------------------------------------------------------------------------------
// Transmit data in serial format
void cmdTx(boolean repeat) {
  char *p = strtok(NULL, "\r"); // I want the full string, parsing is special
  while (*p == ' ') p++;
  if (*p != 0) {
    byte masterBrightness = 255;
    nSerialBuffer = 0;
    while (*p != 0 && nSerialBuffer < sizeof(serialBuffer) - 4) {
      if (*p == '"' || *p == '\'') {   // Quoted string
        p = parseStringInput(p);
      } else if (strchr("RGBWCMYOXIrgbwcmyoxi", *p)) {  // letters for different colors
        char c = toupper(*p++);
        byte v = 255;
        byte r = 0;
        byte g = 0;
        byte b = 0;
        byte w = 0;
        if (isDigit(*p)) v = pixelLevel[byte(*p++) - 0x30];
        switch (c) {
          case 'I':
            masterBrightness = v;
            break;
          case 'R': // Red
            r = v;
            break;
          case 'G': // Green
            g = v;
            break;
          case 'B': // Blue
            b = v;
            break;
          case 'W': // White
            w = v;
            break;
          case 'C': // Cyan
            b = v;
            g = v;
            break;
          case 'M': // Magenta
            r = v;
            b = v;
            break;
          case 'Y': // Yellow
            r = v;
            g = v;
            break;
          case 'O':   // Off
            break;
        }
        if (c != 'I') switch (pixelFormat & 7) {
            case 0: // RGB
              serialBuffer[nSerialBuffer++] = (r + w) * masterBrightness / 255;
              serialBuffer[nSerialBuffer++] = (g + w) * masterBrightness / 255;
              serialBuffer[nSerialBuffer++] = (b + w) * masterBrightness / 255;
              break;
            case 1: // GRB
              serialBuffer[nSerialBuffer++] = (g + w) * masterBrightness / 255;
              serialBuffer[nSerialBuffer++] = (r + w) * masterBrightness / 255;
              serialBuffer[nSerialBuffer++] = (b + w) * masterBrightness / 255;
              break;
            case 2: // RGBW
              serialBuffer[nSerialBuffer++] = r * masterBrightness / 255;
              serialBuffer[nSerialBuffer++] = g * masterBrightness / 255;
              serialBuffer[nSerialBuffer++] = b * masterBrightness / 255;
              serialBuffer[nSerialBuffer++] = w * masterBrightness / 255;
              break;
            case 3: // GRBW
              serialBuffer[nSerialBuffer++] = g * masterBrightness / 255;
              serialBuffer[nSerialBuffer++] = r * masterBrightness / 255;
              serialBuffer[nSerialBuffer++] = b * masterBrightness / 255;
              serialBuffer[nSerialBuffer++] = w * masterBrightness / 255;
              break;
            case 4: // IBGR APA102
              serialBuffer[nSerialBuffer++] = 0xe0 | masterBrightness >> 3; // Led start marker and maximum intensity
              serialBuffer[nSerialBuffer++] = b + w;
              serialBuffer[nSerialBuffer++] = g + w;
              serialBuffer[nSerialBuffer++] = r + w;
              break;
          }
      } else if (*p == '0' && (*(p + 1) == 'x' || *(p + 1) == 'X')) { // Hex number
        p += 2;
        byte n = 0;
        while (*p != 0 && *p != ' ') {
          byte c = *p;
          if (c >= 'a' && c <= 'f') c = c - 'a' + 10;
          else if (c >= 'A' && c <= 'F') c = c - 'A' + 10;
          else if (c >= '0' && c <= '9') c = c - '0';
          else c = 0;
          n = (n << 4) + c;
          p++;
        }
        serialBuffer[nSerialBuffer++] = n;
      } else {  // Number
        char *pp = p;
        while (*p != 0 && *p != ' ') p++;
        if (*p != 0) {
          *p = 0;
          p++;
        }
        serialBuffer[nSerialBuffer++] = (byte)atoi(pp);
      }
      while (*p == ' ') p++;
    }
  } // End of parameter handling
  if (spiFormat & 0x80) setSPI();
  else if (pixelFormat & 0x80) setPixel();
  else setSerial(repeat);
}

void cmdTxQ(Stream * stream) {
  boolean inStr = false;
  for (int i = 0; i < nSerialBuffer; i++) {
    byte c = serialBuffer[i];
    if ((pixelFormat & 0x80) || (c < 32 || c == 127 || c == 0xff)) {
      if (inStr) {
        stream->print("\" ");
        inStr = false;
      }
      stream->print("0x");
      stream->print(c, 16);
      stream->print(' ');
    } else {
      if (!inStr) {
        stream->print("\" ");
        inStr = true;
      }
      stream->print((char) c);
    }

  }
  if (inStr) {
    stream->print("\"");
  }
  stream->print('\n');
}

PROGMEM const char rgbformats[] = "RGB\0PL9823\0TM1804\0\0";
PROGMEM const char grbformats[] = "GRB\0WS2812B\0WS2812\0SK6812\0\0";
PROGMEM const char rgbwformats[] = "RGBW\0SK6812RGBW\0\0";
PROGMEM const char grbwformats[] = "GRBW\0\0";
PROGMEM const char ibgrformats[] = "IBGR#\0APA102#\0SK9822#\0\0";

//----------------------------------------------------------------------------------------------------
// Define serial format
void cmdPixel() {
  spiFormat &= ~0x80;
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && isDigit(*cmd)) {
    pixelCount = atoi(cmd);
    cmd = strtok(NULL, cmdDelimeter);
  }
  if (cmd != NULL) {
    if (cmdidx(cmd, rgbformats) > 0) {
      pixelFormat = 0;
    } else if (cmdidx(cmd, grbformats) > 0) {
      pixelFormat = 1;
    } else if (cmdidx(cmd, rgbwformats) > 0) {
      pixelFormat = 2;
    } else if (cmdidx(cmd, grbwformats) > 0) {
      pixelFormat = 3;
    } else if (cmdidx(cmd, ibgrformats) > 0) {
      pixelFormat = 4;
    }
  }
  pixelFormat |= 0x80;

}

//----------------------------------------------------------------------------------------------------
// Define serial format
void cmdPixelQ(Stream * stream) {
  if (pixelCount > 0) {
    stream->print(pixelCount);
    stream->print(' ');
  }
  switch (pixelFormat & 7) {
    case 0 : stream->print("RGB"); break;
    case 1 : stream->print("GRB"); break;
    case 2 : stream->print("RGBW"); break;
    case 3 : stream->print("GRBW"); break;
    case 4 : stream->print("IBGR#"); break;
  }
  stream->print('\n');
}


//----------------------------------------------------------------------------------------------------
// Define serial format
void cmdSerial() {
  char *cmd = strtok(NULL, cmdDelimeter);
  char *p = strpbrk(cmd, "NSMOECnsmoec");
  spiFormat &= ~0x80;
  pixelFormat &= ~0x80;
  if (p != NULL) {
    switch (toupper(*p)) {
      case 'N': serialFormat = 0 << 5; break;
      case 'S': serialFormat = 1 << 5; break;
      case 'M': serialFormat = 2 << 5; break;
      case 'O': serialFormat = 3 << 5; break;
      case 'E': serialFormat = 4 << 5; break;
      case 'C': serialFormat = 7 << 5; break;
    }
    switch (*(p + 1)) {
      case '5' : serialFormat |= 0; break;
      case '6' : serialFormat |= 1; break;
      case '7' : serialFormat |= 2; break;
      case '8' : serialFormat |= 3; break;
      case '9' : serialFormat |= 3 | (7 << 5); break;
    }
    switch (*(p + 2)) {
      case '1' : serialFormat |= 0x00; break;
      case '2' : serialFormat |= 0x10; break;
    }
    *p = 0;
  }
  if (*cmd != 0) {
    baudrate = atol(cmd);
  }
}

//----------------------------------------------------------------------------------------------------
// Define serial format
void cmdSPI() {
  char *cmd = strtok(NULL, cmdDelimeter);
  char *p = strpbrk(cmd, "CDFRLEcdfrle.");
  spiFormat |= 0x80;
  pixelFormat &= ~0x80;
  if (p != NULL) {
    spiFormat &= ~(0x01 | 0x02 | 0x04 | 0x08); // Start with empty options
    while (*p != 0) {
      switch (toupper(*p)) {
        // C== Falling clock edge
        case 'F':                             // Falling edge with standby high
        case 'C': spiFormat |= 0x01; break;
        case 'D': spiFormat |= 0x02; break;
        case 'R': spiFormat |= 0x03; break;   // Raising edge with standby high
        case 'L': spiFormat |= 0x04; break;   // LSB first
        case 'E': spiFormat |= 0x08; break;   // Use low enable
      }
      *p = 0;
      p++;
    }
  }
  if (*cmd != 0) {
    spirate = atol(cmd);
  }
}



//----------------------------------------------------------------------------------------------------
// Return serial format
void cmdSerialQ(Stream * stream) {
  stream->print(baudrate);
  char c = '?';
  switch ((serialFormat >> 5) & 0x07) {
    case 0: c = 'N'; break;
    case 1: c = 'S'; break;
    case 2: c = 'M'; break;
    case 3: c = 'O'; break;
    case 4: c = 'E'; break;
    case 7: c = 'C'; break;
  }
  stream->print(c);
  stream->print((char) ((serialFormat & 0x07) + '5'));
  stream->print((char) ((serialFormat >> 4 & 0x01) + '1'));
  stream->print('\n');
}


//----------------------------------------------------------------------------------------------------
// Return serial format
void cmdSPIQ(Stream * stream) {
  stream->print(spirate);
  if (spiFormat & 0x01) stream->print('C');
  if (spiFormat & 0x02) stream->print('D');
  if (spiFormat & 0x04) stream->print('L');
  if (spiFormat & 0x08) stream->print('E');
  stream->print('\n');
}


void printTime(Stream * stream, long time) {
  unsigned int v = time / 360000000;
  boolean printed = false;
  if (v > 0) {
    stream->print(v);
    stream->print(":");
    time -= v * 360000000;
    printed = true;
  }
  v = time / 6000000;
  if (v > 0 || printed) {
    stream->print(v);
    stream->print(":");
    time -= v * 6000000;
    printed = true;
  }
  v = time / 100000;
  if (v > 0 || printed) {
    if (time % 100000) stream->print(time / 100000.0, 3); // Some ms present, use decimals
    else stream->print(v);
    if (!printed) stream->print("s"); // Either use : or postfix with s
    printed = true;
  }
  if (!printed) {
    stream->print(time / 100.0, 2);
    stream->print("ms");
  }
}

//----------------------------------------------------------------------------------------------------
// Return last used pulse timers
void cmdPulseQ(Stream * stream) {
  if (mode == MODE_PULSES) {
    printTime(stream, pulseTime);
    stream->print(' ');
    printTime(stream, pulseTimeOff);
    stream->print(' ');
    stream->print(pulseCount);
  } else if (mode == MODE_PULSE || mode == MODE_SLOW_PULSE) {
    printTime(stream, pulseTime);
  } else {
    printTime(stream, pulseT1);
    stream->print(' ');
    printTime(stream, pulseT2);
    stream->print(' ');
    printTime(stream, pulseT3);
  }
  stream->print('\n');
}

//----------------------------------------------------------------------------------------------------
// Return status for a specifik relay output, this ignores pwm and pulse modes.
void cmdRelayQ(Stream * stream) {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    int v = atoi(cmd);
    if (v < 1 || v > 5) return;
    byte stat = mosStatus & (1 << (v - 1));
    stream->print(stat != 0 ? 1 : 0);
    stream->print('\n');
  }
}

//----------------------------------------------------------------------------------------------------
void cmdModeQ(Stream * stream) {
  const char *p = "Unknown";
  switch (mode) {
    case MODE_RELAY : p = "Relay"; break;
    case MODE_PWM : p = "PWM"; break;
    case MODE_PULSE : p = "Pulse"; break;
    case MODE_PULSES : p = "Pulses"; break;
    case MODE_PULSEALL : p = "Pulse all"; break;
    case MODE_SEQ : p = "Sequence"; break;
    case MODE_PPPWM : p = "Push Pull PWM"; break;
    case MODE_PATTERN : p = "Pattern"; break;
    case MODE_SERIAL : p = "Serial"; break;
    case MODE_SPI : p = "SPI"; break;
    case MODE_PIXEL: p = "Pixel"; break;
    case MODE_SLOW_PULSE : p = "Pulse slow"; break;
    case MODE_SLOW_PULSES : p = "Pulses slow"; break;
    case MODE_SLOW_PULSEALL : p = "Pulse all slow"; break;
    case MODE_SLOW_SEQ : p = "Sequence slow"; break;
    case MODE_SLOW_PATTERN : p = "Pattern slow"; break;
  }
  stream->print('"');
  stream->print(p);
  stream->print('"');
  stream->print('\n');
}

//----------------------------------------------------------------------------------------------------
void cmdRegisterQ(Stream * stream) {
  char *cmd = strtok(NULL, cmdDelimeter);
  if (cmd != NULL && *cmd != 0) {
    int v = atoi(cmd);
    byte b = *((byte *) v);
    stream->print(v);
    stream->print(" 0x");
    stream->print(v, 16);
    stream->print(" -> ");
    stream->print(b, 10);
    stream->print(" 0x");
    stream->print(b, 16);
    stream->print(" 0b");
    stream->print(b, 2);
    stream->print('\n');

  }
}

//----------------------------------------------------------------------------------------------------
// Select 1 of 4 program paths, only valid when in program
// The normal button handler is not called when this function is active.
void cmdSelect() {
  int i = 0;
  while (!buttonPressed || buttonStatus == 0) {
    delay(1);
    i++;
    if (i == 70) PORTD = PORTD | (1 << PD1 | 1 << PD0 | 1 << PD4 | 1 << PD7);
    if (i >= 140) {
      PORTD = PORTD & ~(1 << PD1 | 1 << PD0 | 1 << PD4 | 1 << PD7);
      i = 0;
    }
    if (mode > MODE_SLOW_LIMIT) slowHandling();
  }
  findPgm = 0;
  if (buttonStatus & 1) findPgm = '1'; // Search for PGM1
  if (buttonStatus & 2) findPgm = '2'; // Search for PGM2
  if (buttonStatus & 4) findPgm = '3'; // Search for PGM3
  // if (buttonStatus & 8) {} // Continue program, no need to do anything special;
  buttonPressed = false;  // Avoid this button press stops the program
}

//----------------------------------------------------------------------------------------------------
// Return status for buttons
void cmdButtonsQ(Stream * stream) {
  stream->print(buttonStatus);
  stream->print('\n');
}

// Count pulses and turn off when specified number reached
#ifndef TIMER1_COMPA_vect
#error "Undefined interrupt vector"
#endif
ISR(TIMER1_COMPA_vect) {      // interrupt service routine
  pulseCount--;
  if (pulseCount == 0) {
    TCCR1A = 0;
    TCCR1B = 0;
  }
}


//----------------------------------------------------------------------------------------------------
// interrupt service button scanner routine, depending on mode either timer3 or timer4 is used for interrupts
ISR(TIMER4_OVF_vect) {
  byte cnt = (~(PINF >> 4)) & 0x0f;
  if (cnt != buttonStatus) {
    debounce--;
    if (debounce == 0) {
      buttonStatus = cnt;
      buttonPressed = true; // Signal to loop that a button change is detected.
    }
  } else debounce = debounceCount;
}

//----------------------------------------------------------------------------------------------------
// interrupt service button scanner routine, depending on mode either timer3 or timer4 is used for interrupts
ISR(TIMER3_OVF_vect) {
  byte cnt = (~(PINF >> 4)) & 0x0f;
  if (cnt != buttonStatus) {
    debounce--;
    if (debounce == 0) {
      buttonStatus = cnt;
      buttonPressed = true; // Signal to loop that a button change is detected.
    }
  } else debounce = debounceCount;
}

//----------------------------------------------------------------------------------------------------
// Pattern output interrupt
ISR(TIMER1_OVF_vect) {
  byte p = pattern[patternIdx];
  PORTB = (PORTB & ~(0x68)) | (p & 1 ? 0x20 : 0) | (p & 2 ? 0x40 : 0) | (p & 8 ? 0x08 : 0);
  PORTC = (PORTC & ~(0x40)) | (p & 4 ? 0x40 : 0);

  if (repeatPattern) patternIdx = (patternIdx + 1) % npattern;
  else if (patternIdx < npattern - 1) {
    patternIdx++;
  }
}

//----------------------------------------------------------------------------------------------------
// Serial interupts
#if defined(UART1_UDRE_vect)
ISR(UART1_UDRE_vect)
#elif defined(USART1_UDRE_vect)
ISR(USART1_UDRE_vect)
#else
#error "Don't know what the Data Register Empty vector is called for Serial1"
#endif
{
  byte b = serialBuffer[serialBufferIdx];
  if (serialBufferParityMask == 0) {
    UCSR1B &= ~(1 << TXB81);
  } else if (serialBufferParityMask & 0xf0) { // Mark parity in bit 7 or 6
    b |= (serialBufferParityMask & 0xf0);
  } else if (serialBufferParityMask == 0x03) {
    UCSR1B |= (1 << TXB81); // Setup bit 8 as mark for command
    serialBufferParityMask = 0; // The rest of the data is with a 0 bit 8
  } else if (serialBufferParityMask & 0x02) { // Same as above but in bit 7 or 6
    b |= (serialBufferParityMask & 0xf0);
    serialBufferParityMask = 0; // The rest of the data is with a 0  in this bit
  } else {
    UCSR1B |= (1 << TXB81); // Setup bit 8 as 1 for mark parity
  }
  PORTB |= (1 << PB6);             // Signal transmission starts now
  UDR1 = b;
  if (repeatSerialBuffer) serialBufferIdx = (byte) (serialBufferIdx + 1) % nSerialBuffer;
  else if (serialBufferIdx < nSerialBuffer - 1) {
    serialBufferIdx++;
  } else {
    UCSR1B &= ~(1 << UDRIE1);    // Turn interrupt off
  }
}

//----------------------------------------------------------------------------------------------------
#if defined(UART1_TX_vect)
ISR(UART1_TX_vect)
#elif defined(USART1_UDRE_vect)
ISR(USART1_TX_vect)
#else
#error "Don't know what the Data Register Empty vector is called for Serial1"
#endif
{
  PORTB &= ~(1 << PB6);       // Turn data transmitting pin off
  UCSR1B &= ~(1 << TXCIE1);    // Turn interrupt off
}

//----------------------------------------------------------------------------------------------------
#if defined(SPI_STC_vect)
ISR(SPI_STC_vect)
#else
#error "Don't know what the vector is called for SPI"
#endif
{
  if (serialBufferIdx < nSerialBuffer) {
    SPDR = serialBuffer[serialBufferIdx];
    serialBufferIdx++;
  } else {
    SPCR &= ~(1 << SPIE);   // Turn interrrupt off
    if (spiFormat & 0x08) PORTB |= (1 << PB6); // Turn data transmitting pin off
    else PORTB &= ~(1 << PB6);
  }
}

//----------------------------------------------------------------------------------------------------
void slowHandling() {
  if (millis() >= nextSlowEvent) {
    switch (mode) {
      case MODE_SLOW_PATTERN :
        updateSlowPattern();
        break;
      case MODE_SLOW_PULSE :
        updateSlowPulse();
        break;
      case MODE_SLOW_PULSES :
        updateSlowPulses();
        break;
      case MODE_SLOW_PULSEALL :
        updateSlowPulseAll();
        break;
      case MODE_SLOW_SEQ :
        updateSlowSequence();
        break;
    }
  }
}


//----------------------------------------------------------------------------------------------------
void handleButtonPressed() {
  if (buttonStatus == 0) return; // All buttons released
  switch (mode) {
    case MODE_RELAY:      // Toggle relay output
      mosStatus ^= (buttonStatus & 0x07);
      setRelays();
      break;
    case MODE_PWM:        // Set duty cycle to 0
      if (buttonStatus & 0x01) {
        pwmDutyCycle1 = 0;
      }
      if (buttonStatus & 0x02) {
        pwmDutyCycle2 = 0;
      }
      if (buttonStatus & 0x04) {
        pwmDutyCycle3 = 0;
      }
      if (pwmDutyCycle1 == 0 && pwmDutyCycle2 == 0 && pwmDutyCycle3 == 0) {
        setRelays();
      } else {
        setPWM();
      }
      break;
    default:
      if (buttonStatus & 0x07) {    // Change to relay mode
        setRelays();
      }
      break;
  }
  if (buttonStatus & 0x08 && !inProgram && !loopProgram) {    // This button will run a stored program
    cmdRun(NULL);
  }
}

//----------------------------------------------------------------------------------------------------
void loop() {
  setLeds();
  findPgm = 0;          // Stop any aborts and searching.
  if (loopProgram || findPgm > '0') cmdRun(NULL);
  if (!loopProgram) inProgram = false;
  if (mode > MODE_SLOW_LIMIT) slowHandling();

  cmdProcessor.processInput();
  if (buttonPressed) {
    buttonPressed = false;
    loopProgram = false;
    handleButtonPressed();
  }
}
